@@ -36,6 +36,7 @@
 #include "mongoc-stream.h"
 #include "mongoc-uri.h"
 #include "mongoc-write-concern.h"
+#include "mongoc-read-concern.h"
 
 
 BSON_BEGIN_DECLS
@@ -140,6 +141,9 @@ int32_t                        mongoc_client_get_max_bson_size    (mongoc_client
 const mongoc_write_concern_t  *mongoc_client_get_write_concern    (const mongoc_client_t        *client);
 void                           mongoc_client_set_write_concern    (mongoc_client_t              *client,
                                                                    const mongoc_write_concern_t *write_concern);
+const mongoc_read_concern_t   *mongoc_client_get_read_concern     (const mongoc_client_t        *client);
+void                           mongoc_client_set_read_concern     (mongoc_client_t              *client,
+                                                                   const mongoc_read_concern_t  *read_concern);
 const mongoc_read_prefs_t     *mongoc_client_get_read_prefs       (const mongoc_client_t        *client);
 void                           mongoc_client_set_read_prefs       (mongoc_client_t              *client,
                                                                    const mongoc_read_prefs_t    *read_prefs);
